.jp-git-diff-error {
  width: auto;
  padding: 16px;
  color: var(--jp-ui-font-color1);
  font-size: var(--jp-ui-font-size1);
  background: var(--jp-error-color3);
}

.jp-git-diff-error-message {
  color: var(--jp-error-color1);
  margin-left: 5px;
}

.jp-git-diff-root {
  min-width: 2px;
  min-height: 30px;
  height: 100%;
  display: flex;
  flex-direction: column;
  outline: none;
  background: var(--jp-layout-color0);
  color: var(--jp-ui-font-color0);
}

button.jp-git-diff-refresh {
  border: solid 1px var(--jp-warn-color2);
}

button.jp-git-diff-resolve {
  border: solid 1px var(--jp-success-color2);
}

button.jp-git-diff-refresh .jp-ToolbarButtonComponent-label {
  color: var(--jp-warn-color1);
}

button.jp-git-diff-resolve .jp-ToolbarButtonComponent-label {
  color: var(--jp-success-color1);
}

.jp-git-diff-banner, .jp-git-merge-banner {
  flex: 0 0 auto;
  margin: 2px 0px;
  display: flex;
}

.jp-git-diff-banner span{
  padding: 0px 4px;
  width: 47%;
  font-weight: bold;
}

.jp-git-diff-banner span.jp-spacer {
  width: 6%;
}

.jp-git-diff-banner span:first-of-type {
  background-color: var(--jp-git-diff-deleted-color);
}

.jp-git-diff-banner span:last-of-type {
  background-color: var(--jp-git-diff-added-color);
}

.jp-git-merge-banner span {
  padding: 0px 4px;
  width: 31%;
  font-weight: bold;
}

.jp-git-merge-banner span.jp-spacer {
  width: 3.5%;
}

.jp-git-diff-root .CodeMirror-merge {
  position: relative;
  white-space: pre;
  border: var(--codemirror-border);
  border-radius: 0px;
  background: var(--jp-cell-editor-background);
}

.jp-git-diff-root .CodeMirror {
  background: var(--jp-layout-color0);
}

.jp-git-diff-root .CodeMirror-merge,
.CodeMirror-merge .CodeMirror {
  height: auto;
}

.jp-git-diff-root .CodeMirror-merge-4pane .CodeMirror-merge-pane-deleted {
  display: none;
}

.jp-git-diff-root .CodeMirror-merge-pane-unchanged {
  width: 100%;
}

.jp-git-diff-root .CodeMirror-merge-1pane .CodeMirror-merge-gap {
  width: 6%;
}

.jp-git-diff-root .CodeMirror-merge-2pane .CodeMirror-merge-pane {
  width: 47%;
}

.jp-git-diff-root .CodeMirror-merge-2pane .CodeMirror-merge-gap {
  width: 6%;
}

.jp-git-diff-root .CodeMirror-merge-3pane .CodeMirror-merge-pane {
  width: 31%;
}

.jp-git-diff-root .CodeMirror-merge-3pane .CodeMirror-merge-gap {
  width: 3.5%;
}

.jp-git-diff-root .CodeMirror-merge-pane {
  display: inline-block;
  white-space: normal;
  vertical-align: top;
}

.jp-git-diff-root .CodeMirror-merge-pane-rightmost {
  position: absolute;
  right: 0px;
  z-index: 1;
}

.jp-git-diff-root .CodeMirror-merge-scrolllock-wrap {
  position: absolute;
  bottom: 0;
  left: 50%;
}

.jp-git-diff-root .CodeMirror-merge-scrolllock {
  position: relative;
  left: -50%;
  cursor: pointer;
  line-height: 1;
}

.jp-git-diff-root .CodeMirror-merge-r-inserted,
.jp-git-diff-root .CodeMirror-merge-l-inserted {
  background-color: var(--jp-git-diff-added-color);
}

.jp-git-diff-root .CodeMirror-merge-r-deleted,
.jp-git-diff-root .CodeMirror-merge-l-deleted {
  background-color: var(--jp-git-diff-deleted-color);
}

.jp-git-diff-root .CodeMirror-merge-collapsed-widget:before {
  content: '(...)';
}

.jp-git-diff-root .CodeMirror-merge-collapsed-widget {
  cursor: pointer;
  color: var(--jp-ui-font-color1);
  background: var(--jp-layout-color2);
  border: 1px solid var(--jp-border-color1);
  font-size: 90%;
  padding: 0 3px;
  border-radius: 4px;
}

.jp-git-diff-root .CodeMirror-merge-collapsed-line .CodeMirror-gutter-elt {
  display: none;
}

.jp-git-diff-root .CodeMirror-merge-spacer {
  background-image: repeating-linear-gradient(
    145deg,
    transparent,
    transparent 10px,
    var(--jp-border-color0) 10px,
    var(--jp-border-color0) 12px
  );
}
