.jp-CredentialsBox input[type='text'],
.jp-CredentialsBox input[type='password'] {
  display: block;
  width: 100%;
  margin-top: 10px;
  margin-bottom: 10px;
}

.jp-CredentialsBox input[type='checkbox'] {
  display: inline-block;
}

.jp-CredentialsBox-warning {
  color: var(--jp-warn-color0);
}

.jp-CredentialsBox-label-checkbox {
  display: flex;
  align-items: center;
}
